import os
from glob import glob
import numpy as np
import h5py


class H5TTable(object):
    """docstring for H5TTable"""

    def __init__(self, path, dset_key=None):
        self.path = path
        self.hf = h5py.File(path, 'r')
        self.keys = list(self.hf.keys())

        self.dset = None

        if dset_key is not None:
            self.set_dataset(dset_key)

        self.sites = self.hf['sites'][:].astype('U6')
        self.stations = self.hf['stations'][:].astype('U4')
        self.station_locations = self.hf['station_locations'].astype('U2')
        self._sitedict = dict(zip(self.sites, np.arange(len(self.sites))))

        self.locations = self.hf['locations'][:]
        self.coords = self.hf['grid_locs'][:]

    def __delete__(self):
        sefl.hf.close()

    def set_dataset(self, key):
        if key in self.keys:
            self.dset = self.hf[key]
        else:
            raise KeyError('dataset %s does not exist' % key)

    @property
    def shape(self):
        return self.hf.attrs['shape']

    @property
    def origin(self):
        return self.hf.attrs['origin']

    @property
    def spacing(self):
        return self.hf.attrs['spacing']

    def index_sites(self, sites):
        if isinstance(sites, (list, np.ndarray)):
            return np.array([self._sitedict[site] for site in sites])
        else:
            return self._sitedict[site]

    def icol_to_xyz(self, index):
        nx, ny, nz = self.shape
        iz = index % nz
        iy = ((index - iz) // nz) % ny
        ix = index // (nz * ny)
        loc = np.array([ix, iy, iz], dtype=float) * self.spacing + self.origin

        return loc

    def xyz_to_icol(self, loc):
        x, y, z = loc
        ix, iy, iz = ((loc - self.origin) / self.spacing).astype(int)
        nx, ny, nz = self.shape
        # return (iz * nx * ny) + (iy * nx) + ix;

        return int((ix * ny * nz) + (iy * nz) + iz)

    def close(self):
        self.hf.close()


def gdef_to_points(shape, origin, spacing):
    maxes = origin + shape * spacing
    x = np.arange(origin[0], maxes[0], spacing[0]).astype(np.float32)
    y = np.arange(origin[1], maxes[1], spacing[1]).astype(np.float32)
    z = np.arange(origin[2], maxes[2], spacing[2]).astype(np.float32)
    points = np.zeros((np.product(shape), 3), dtype=np.float32)
    ix = 0

    for xv in x:
        for yv in y:
            for zv in z:
                points[ix] = [xv, yv, zv]
                ix += 1

    return points


def array_from_travel_time_ensemble(tt_grids):

    data = {'P': [],
            'S': []}

    sites = []
    slocs = []
    shape = tt_grids[0].shape
    origin = tt_grids[0].origin
    spacing = tt_grids[0].spacing
    for tt_grid in tt_grids:
        sites.append(tt_grid.seed_label)
        slocs.append(tt_grid.seed)

    sites = np.array(sites)
    slocs = np.array(slocs)

    nsites = len(sites)
    ngrid = np.product(shape)

    tts = np.zeros((nsites, ngrid), dtype=np.float32)

    for i in range(nsites):
        tts[i] = tt_grids[i].data.reshape(ngrid).astype(np.float32)

    data[phase] = dict(ttable=tts, locations=slocs, shape=shape,
                       origin=origin, spacing=spacing, sites=sites)

    return data


def write_hdf5(fname, tt_grids):

    hf = h5py.File(fname, 'w')

    shape = tt_grids.shape
    spacing = tt_grids.spacing
    origin = tt_grids.origin

    hf.attrs['shape'] = shape
    hf.attrs['spacing'] = spacing
    hf.attrs['origin'] = origin

    sites = tt_grids.seed_labels
    locations = tt_grids.seeds

    hf.create_dataset('locations', data=locations.astype(np.float32))
    gridlocs = gdef_to_points(shape, origin, spacing)
    hf.create_dataset('grid_locs', data=gridlocs.astype(np.float32))
    gdef = np.concatenate((shape, origin, spacing)).astype(np.int32)
    hf.create_dataset('grid_def', data=gdef)
    hf.create_dataset('sites', data=sites.astype('S8'))
    stations = np.array([site[0:6] for site in sites])
    station_locations = np.array([site[6:] for site in sites])
    hf.create_dataset('stations', data=stations.astype('S6'))
    hf.create_dataset('station_locations', data=station_locations.astype('S2'))

    nsites = len(sites)
    ngrid = np.product(shape)

    tts = {'P': np.zeros((nsites, ngrid), dtype=np.float32),
           'S': np.zeros((nsites, ngrid), dtype=np.float32)}

    for i, site in enumerate(sites):
        for phase in ['P', 'S']:
            tt_grid = tt_grids.select(phase=phase, seed_labels=site)[0]
            tts[phase][i] = tt_grid.data.reshape(ngrid).astype(np.float32)

    hf.create_dataset('ttp', data=tts['P'])
    hf.create_dataset('tts', data=tts['S'])
    hf.close()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: core.py
#  Purpose: Expansion of the obspy.core.trace module
#   Author: uquake development team
#    Email: devs@uquake.org
#
# Copyright (C) 2016 uquake development team
# --------------------------------------------------------------------
"""
Expansion of the obspy.core.trace module

:copyright:
    uquake development team (devs@uquake.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
from abc import ABC
import numpy as np
from obspy.core.trace import Trace as ObspyTrace
from obspy.core.trace import Stats as ObspyStats
from obspy import UTCDateTime
from obspy.core.event import WaveformStreamID
from obspy.core.trace import AttribDict

from .event import Pick
from .util import tools


class Stats(ObspyStats, ABC):
    __doc__ = ObspyStats.__doc__.replace('obspy', 'uquake')

    def __init__(self, stats=None, **kwargs):
        super(Stats, self).__init__(**kwargs)

        if stats:
            for item in stats.__dict__.keys():
                self.__dict__[item] = stats.__dict__[item]

            self.site = f'{self.station}{self.location}'

    def __str__(self):
        """
        Return better readable string representation of Stats object.
        """
        priorized_keys = ['network', 'station', 'site', 'location', 'channel',
                          'starttime', 'endtime', 'sampling_rate', 'delta',
                          'npts', 'calib']
        return self._pretty_str(priorized_keys)


# class Trace(obsstream.Trace, ABC):
#     __doc__ = obsstream.Trace.__doc__.replace('obspy', 'uquake')
#
#     def __init__(self, trace=None, **kwargs):
#         super(Trace, self).__init__(**kwargs)
#
#         if trace:
#             self.stats = Stats(stats=trace.stats)
#             self.data = trace.data
#
#         if 'header' in kwargs.keys():
#             self.stats = Stats(stats=kwargs['header'])


class Trace(ObspyTrace, ABC):
    def __init__(self, trace=None, **kwargs):
        super(Trace, self).__init__(**kwargs)

        if trace:
            self.stats = Stats(stats=trace.stats)
            self.data = trace.data

        if 'header' in kwargs.keys():
            self.stats = Stats(stats=kwargs['header'])

        if 'stats' in kwargs.keys():
            self.stats = Stats(stats=kwargs['stats'])

    @property
    def sr(self):
        return self.stats.sampling_rate

    @property
    def ppv(self):
        return np.max(np.abs(self.data))

    @property
    def ppa(self):
        return None

    def time_to_index(self, time):
        return int((time - self.stats.starttime) * self.sr)

    def index_to_time(self, index):
        return self.stats.starttime + (index / self.sr)

    def times(self):
        sr = self.stats.sampling_rate

        return np.linspace(0, len(self.data) / sr, len(self.data))

    def plot(self, **kwargs):
        from uquake.imaging.waveform import WaveformPlotting
        waveform = WaveformPlotting(stream=self, **kwargs)

        return waveform.plotWaveform()

    def make_pick(self, pick_time, wlen_search,
                  stepsize, snr_wlens, phase_hint=None):

        ipick = self.time_to_index(pick_time)
        sr = self.stats.sampling_rate
        stepsize_samp = int(stepsize * sr)
        snr_wlens_samp = (snr_wlens * sr).astype(int)
        wlen_search_samp = int(wlen_search * sr)

        new_pick, snr = tools.repick_using_snr(self.data, ipick,
                                               wlen_search_samp, stepsize_samp,
                                               snr_wlens_samp)

        waveform_id = WaveformStreamID(channel_code=self.stats.channel,
                                       station_code=self.stats.station)

        pick = Pick(time=self.index_to_time(newpick), waveform_id=waveform_id,
                    phase_hint=phase_hint, evaluation_mode='automatic',
                    evaluation_status='preliminary', method='snr', snr=snr)

        return pick

    def time_within(self, utime, edge_buf=0.0):
        within = True

        if (utime - edge_buf) < self.stats.starttime:
            within = False
        elif (utime + edge_buf) > self.stats.endtime:
            within = False

        return within

    @classmethod
    def create_from_json(cls, trace_json_object):

        trace_json_object['stats']['starttime'] = UTCDateTime(
            trace_json_object['stats']['starttime'])
        trace_json_object['stats']['endtime'] = UTCDateTime(
            trace_json_object['stats']['endtime'])

        trc = cls(header=trace_json_object['stats'], data=np.array(
            trace_json_object['data'], dtype='float32'))

        return trc

    def to_json(self):
        trace_dict = dict()
        trace_dict['stats'] = dict()

        for key in self.stats.keys():
            if isinstance(self.stats[key], UTCDateTime):
                trace_dict['stats'][key] = self.stats[key].isoformat()
            elif isinstance(self.stats[key], AttribDict):
                trace_dict['stats'][key] = self.stats[key].__dict__
            else:
                trace_dict['stats'][key] = self.stats[key]

        trace_dict['data'] = self.data.tolist()

        return trace_dict



from ...core.inventory import (Inventory, Network, Station, Site, Channel,
                                   geophone_response, SystemResponse)
from ...core import UTCDateTime
from datetime import datetime
from typing import TextIO, Union, List, Dict
from pathlib import Path
from pandas import DataFrame
import string
import random


# The naming convention in the sensor file is not consistent, making it hard
# to translate the .sen file into a more
# standard representation

# The below table is not great
sensor_type_lookup_table = {
    'G3': 'geophone',
    'G1': 'geophone',
    'S3': 'geophone',
    'S1': 'geophone'
}

channel_code_lookup_table = {
    'geophone': 'GP'
}


def create_sensor_dictionary(column_name_line: List, sensor_lines: List) \
        -> DataFrame:
    columns = column_name_line.split()
    sensor_dict = {}

    # initializing the dictionary with the header
    columns[0] = columns[0][2:]
    new_columns = []
    for column in columns:
        if column == 'Orientation':
            new_columns.append('Orientation_N')
            new_columns.append('Orientation_E')
            new_columns.append('Orientation_U')
        else:
            new_columns.append(column)
    columns = new_columns
    for column in columns:
        sensor_dict[column] = []

    sensor_dict['location_code'] = []
    sensor_dict['station_code'] = []

    # filling the dictionary list
    for sensor_line in sensor_lines:
        line_elements = sensor_line.split()
        for i, column in enumerate(columns):
            if column == 'Description':
                try:
                    location_code = int(line_elements[i][-3:] - 1)
                    if location_code < 0:
                        location_code = 0
                    sensor_dict['location_code'].append(location_code)
                    sensor_dict['station_code'].append(line_elements[i][:-4])
                except Exception as e:
                    sensor_dict['location_code'].append(0)
                    sensor_dict['station_code'].append(line_elements[i])

            try:
                line_element = float(line_elements[i])
            except Exception as e:
                line_element = line_elements[i]
            sensor_dict[column].append(line_element)

    return DataFrame(sensor_dict)


def create_channels(sensor_df: DataFrame) -> Dict:
    dict_out = {}
    for row in sensor_df.iterrows():
        row = row[1]
        channel_code = f'{channel_code_lookup_table[sensor_type_lookup_table[row["Type"]]]}{row["Cmp"]:0.0f}'
        location_code = f'{row["location_code"]:02d}'

        # Important Notes:
        # Sampling rate is assumed and given a value of 6000, this information
        # should be available
        # start_date is not defined in the sensor file but is a value that is
        # important
        # end_date is not defined in the sensor file but is a value that is
        # important if the sensor has been decommissioned.

        channel = Channel(code=channel_code, location_code=location_code,
                          x=row['Easting'], y=row['Northing'], z=row['Elev'],
                          sample_rate=6000,
                          start_date=UTCDateTime(datetime.fromtimestamp(0)))
        channel.set_orientation([row['Orientation_E'], row['Orientation_N'],
                                 row['Orientation_U']])
        if sensor_type_lookup_table[row['Type']] == 'geophone':
            # assuming for now that the sensor is critically damped

            system_response = SystemResponse()

            ### Need to handle the gain row['Gain']

            system_response.add_geophone(row['Frequency'], row['Sens'])
            system_response.add_digitizer(max_voltage=row['Vmax'],
                                          digitizer_name='Paladin')

            sensitivity = 1
            for stage in system_response.response_stages:
                sensitivity *= stage.stage_gain

            system_response.sensitivity = sensitivity

            channel.response = system_response.response

            # channel.response = geophone_response(row['Frequency'],
            # row['Gain'], sensitivity=row['Sens'])
            # There might be some change required to convert the data in the
            # waveform files that are expressed in ADC counts. The previous
            # and following line may not be accurate.
            channel.calibration_units = 'COUNT'

            # To be rigorous, the sensor, pre-amplifier, digitizer should be
            # added
            # channel.sensor, channel.pre_amplifier, channel.data_logger.

        if row['station_code'] not in dict_out.keys():
            dict_out[row['station_code']] = []

        dict_out[row['station_code']].append(channel)

    return dict_out


def create_station(channel_dict: Dict):
    stations = []
    for key in channel_dict.keys():
        letters = string.ascii_uppercase
        station_code = ''.join(random.choice(letters) for i in range(4))
        stations.append(Station(code=station_code, channels=channel_dict[key],
                                total_number_of_channels=
                                len(channel_dict[key]),
                                selected_number_of_channels=
                                len(channel_dict[key]),
                                start_date=
                                UTCDateTime(datetime.fromtimestamp(0)),
                                alternate_code=key))
    return stations


def read_esg_sensor_file(sen_file: Union[TextIO, str, Path], network: str = "",
                         header_length: int = 38) -> Inventory:
    if not isinstance(sen_file, TextIO):
        sen_file = open(sen_file)

    # Reading the file and removing the header
    previous_site = None

    sensor_file_lines = sen_file.readlines()
    column_name_line = sensor_file_lines[header_length - 2]
    sensor_lines = sensor_file_lines[header_length:]
    sensor_df = create_sensor_dictionary(column_name_line, sensor_lines)

    dict_channel = create_channels(sensor_df)
    stations = create_station(dict_channel)
    network = Network(code=network, stations=stations)
    inventory = Inventory(networks=[network], source='ESG Solutions',
                          sender='ESG Solutions', module='uQuake',
                          module_uri="")
    return inventory


# if __name__ == '__main__':
#     inventory = read_esg_sensor_file("esg_test.sen", network='GB')
#     inventory.write('ESG_test.xml', format='stationxml')



from datetime import datetime, timedelta

import numpy as np
from scipy.fftpack import fft, fftfreq, ifft
from scipy.signal import iirfilter, sosfilt, zpk2sos
import pyproj


def lon_lat_x_y(input_projection, output_projection, longitude, latitude):
    transformer = pyproj.Transformer.from_crs(input_projection,
                                              output_projection,
                                              always_xy=True)

    return transformer.transform(longitude, latitude)


def datetime_to_epoch_sec(dtime):
    return (dtime - datetime(1970, 1, 1)) / timedelta(seconds=1)


def make_picks(stcomp, pick_times_utc, phase, pick_params):
    snr_wlens = np.array(pick_params.snr_wlens)
    wlen_search = pick_params.wlen_search
    stepsize = pick_params.stepsize
    edge_time = wlen_search / 2 + np.max(snr_wlens)

    picks = []

    for tr, ptime in zip(stcomp, pick_times_utc):

        if tr.time_within(ptime, edge_time) is True:
            picks.append(tr.make_pick(ptime, wlen_search,
                                      stepsize, snr_wlens, phase_hint=phase))

    return picks


def copy_picks_to_dict(picks):
    pick_dict = {}

    for pick in picks:
        station = pick.waveform_id.station_code
        phase = pick.phase_hint

        if station not in pick_dict:
            pick_dict[station] = {}
        # MTH: If you copy the pick you pollute the reference id space
        #      and arrival.pick_id.get_referred_object() no longer works!
        # pick_dict[station][phase]=copy.deepcopy(pick)
        pick_dict[station][phase] = pick

    return pick_dict


def picks_to_dict(picks):
    pd = {}

    for p in picks:
        key = p.waveform_id.get_seed_string()

        if key not in pd:
            pd[key] = []
        pd[key].append(p.time)

    return pd


def repick_using_snr(sig, ipick, wlen_search, stepsize, snr_wlens):
    origin_inds, snrs = sliding_snr(sig, ipick, wlen_search, stepsize,
                                    snr_wlens)
    newpick = origin_inds[np.argmax(snrs)]
    snr = np.max(snrs)

    return newpick, snr


def sliding_snr(sig, ipick, wlen_search, stepsize, snr_wlens, plot=False):
    wl_noise, wl_sig = snr_wlens.astype(int)
    hl = int(wlen_search // 2)
    i0 = max(wl_noise, ipick - hl)
    i1 = min(len(sig) - wl_sig, ipick + hl)

    origin_inds = np.arange(i0, i1, stepsize)
    snrs = np.zeros(len(origin_inds), dtype=np.float32)

    for i, og in enumerate(origin_inds):
        energy_noise = np.mean((sig[og - wl_noise:og]) ** 2)
        energy_sig = np.mean((sig[og:og + wl_sig]) ** 2)
        snrs[i] = energy_sig / energy_noise
    snrs = 10 * np.log10(snrs)

    if plot:
        import matplotlib.pyplot as plt
        plt.plot(sig / np.max(sig))
        plt.plot(origin_inds, snrs)
        plt.axvline(ipick, color='green', label='pick_old')
        plt.axvline(i0, color='black', label='scan_win')
        plt.axvline(i1, color='black')

        ot = origin_inds[np.argmax(snrs)]
        plt.axvline(ot, color='green', linestyle='--', label='pick_new')
        plt.axvline(ot - wl_noise, color='red', label='snr_win')
        plt.axvline(ot + wl_sig, color='red')
        plt.legend()

    return origin_inds, snrs


def create_composite(sigs, groups):
    nsig = len(groups)
    npts = sigs.shape[1]

    out = np.zeros((nsig, npts), dtype=sigs.dtype)

    for i, group in enumerate(groups):
        out[i] = np.sign(sigs[group[0]]) * np.sqrt(
            np.sum(sigs[group] ** 2, axis=0))

    return out


def stream_to_array(st, t0, npts_fix, taplen=0.05):
    sr = st[0].stats.sampling_rate
    nsig = len(st)
    taplen_npts = int(npts_fix * taplen)

    data = np.zeros((nsig, npts_fix), dtype=np.float32)

    for i, tr in enumerate(st):
        i0 = int((tr.stats.starttime - t0) * sr + 0.5)
        sig = tr.data - np.mean(tr.data)

        if taplen != 0:
            taper_data(sig, taplen_npts)
        slen = min(len(sig), npts_fix - i0)
        data[i, i0: i0 + slen] = sig[:slen]

    return data


def taper_data(data, wlen):
    tap = hann_half(wlen)
    data[:wlen] *= tap
    data[-wlen:] *= tap[::-1]


def taper2d(data, wlen):
    out = data.copy()
    tap = hann_half(wlen)

    for i in range(data.shape[0]):
        out[i][:wlen] *= tap
        out[i][-wlen:] *= tap[::-1]

    return out


def ttsamp(dist, vel, sr):
    return int(dist / vel * sr + 0.5)


def integrate(sig):
    from scipy.integrate import cumtrapz

    return cumtrapz(sig, initial=0)


def attenuate(sig, sr, dist, Q, vel, gspread=True):
    npts = len(sig)
    fsig = fft(sig)
    freqs = fftfreq(npts, d=1. / sr)
    tstar = dist / (vel * Q)
    factor = np.exp(-np.pi * np.abs(freqs) * tstar)
    fsig *= factor
    sig = np.real(ifft(fsig))

    if gspread:
        sig /= dist

    return sig


def roll_data(data, tts):
    droll = np.zeros_like(data)

    for i, sig in enumerate(data):
        droll[i] = np.roll(sig, -tts[i])

    return droll


def velstack(data, dists2src, sr, vels):
    dnorm = norm2d(data)
    dstack = np.zeros((len(vels), dnorm.shape[1]), dtype=np.float32)

    for ivel, vel in enumerate(vels):
        shifts = (dists2src / vel * sr + 0.5).astype(int)

        for i, shift in enumerate(shifts):
            dstack[ivel] += np.roll(dnorm[i], -shift)

    return dstack


def chan_groups(chanmap):
    return [np.where(sk == chanmap)[0] for sk in np.unique(chanmap)]


def comb_channels(data, cmap):
    groups = [np.where(sk == cmap)[0] for sk in np.unique(cmap)]
    dstack = np.zeros((len(groups), data.shape[1]))

    for i, grp in enumerate(groups):
        dstack[i] = np.mean(np.abs(data[grp]), axis=0)

    return dstack


def bandpass(data, band, sr, corners=4, zerophase=True):
    freqmin, freqmax = band
    fe = 0.5 * sr
    low = freqmin / fe
    high = freqmax / fe

    z, p, k = iirfilter(corners, [low, high], btype='band',
                        ftype='butter', output='zpk')
    sos = zpk2sos(z, p, k)

    if zerophase:
        firstpass = sosfilt(sos, data)

        if len(data.shape) == 1:
            return sosfilt(sos, firstpass[::-1])[::-1]
        else:
            return np.fliplr(sosfilt(sos, np.fliplr(firstpass)))
    else:
        return sosfilt(sos, data)


def filter(data, btype, band, sr, corners=4, zerophase=True):
    # btype: lowpass, highpass, band

    fe = 0.5 * sr
    z, p, k = iirfilter(corners, band / fe, btype=btype,
                        ftype='butter', output='zpk')
    sos = zpk2sos(z, p, k)

    if zerophase:
        firstpass = sosfilt(sos, data)

        if len(data.shape) == 1:
            return sosfilt(sos, firstpass[::-1])[::-1]
        else:
            return np.fliplr(sosfilt(sos, np.fliplr(firstpass)))
    else:
        return sosfilt(sos, data)


def decimate(data_in, sr, factor):
    data = data_in.copy()
    fmax = sr / (factor * 2)
    filter(data, 'lowpass', fmax, sr)

    if len(data.shape) == 1:
        return data[::factor]
    else:
        return data[:, ::factor]


def norm2d(d):
    return d / np.max(np.abs(d), axis=1)[:, np.newaxis]


def cross_corr(sig1, sig2, norm=True, pad=False, phase_only=False, phat=False):
    """Cross-correlate two signals."""
    pad_len = len(sig1)

    if pad is True:
        pad_len *= 2
        # pad_len = signal.next_pow_2(pad_len)

    sig1f = fft(sig1, pad_len)
    sig2f = fft(sig2, pad_len)

    if phase_only is True:
        ccf = np.exp(- 1j * np.angle(sig1f)) * np.exp(1j * np.angle(sig2f))
    else:
        ccf = np.conj(sig1f) * sig2f

    if phat:
        ccf = ccf / np.abs(ccf)

    cc = np.real(ifft(ccf))

    if norm:
        cc /= np.sqrt(energy(sig1) * energy(sig2))

    return np.roll(cc, len(cc) // 2)


def energy(sig, axis=None):
    return np.sum(sig ** 2, axis=axis)


def energy_freq(fsig, axis=None):
    return np.sum(np.abs(fsig) ** 2, axis=axis) / fsig.shape[-1]


def freq_window(cf, npts, sr):
    nfreq = int(npts // 2 + 1)
    fsr = npts / sr
    cf = np.array(cf, dtype=float)
    cx = (cf * fsr + 0.5).astype(int)

    win = np.zeros(nfreq, dtype=np.float32)
    win[:cx[0]] = 0
    win[cx[0]:cx[1]] = taper_cosine(cx[1] - cx[0])
    win[cx[1]:cx[2]] = 1
    win[cx[2]:cx[3]] = taper_cosine(cx[3] - cx[2])[::-1]
    win[cx[-1]:] = 0

    return win


def taper_cosine(wlen):
    return np.cos(np.linspace(np.pi / 2., np.pi, wlen)) ** 2


def phase(sig):
    return np.exp(1j * np.angle(sig))


def whiten2D(a, freqs, sr):
    wl = a.shape[1]
    win = freq_window(freqs, wl, sr)
    af = fft(a)

    for sx in range(a.shape[0]):
        whiten_freq(af[sx], win)
    a[:] = np.real(ifft(af))


def whiten(sig, win):
    """Whiten signal, modified from MSNoise."""
    npts = len(sig)
    nfreq = int(npts // 2 + 1)

    assert (len(win) == nfreq)
    # fsr = npts / sr

    fsig = fft(sig)
    # hl = nfreq // 2

    half = fsig[: nfreq]
    half = win * phase(half)
    fsig[: nfreq] = half
    fsig[-nfreq + 1:] = half[1:].conjugate()[::-1]

    return np.real(ifft(fsig))


def whiten_freq(fsig, win):
    # npts = len(fsig)
    # nfreq = int(npts // 2 + 1)
    nfreq = int(len(fsig) // 2 + 1)
    assert (len(win) == nfreq)
    fsig[: nfreq] = win * phase(fsig[: nfreq])
    fsig[-nfreq + 1:] = fsig[1: nfreq].conjugate()[::-1]


def mirror_freqs(data):
    nfreq = int(data.shape[1] // 2 + 1)
    data[:, -nfreq + 1:] = np.fliplr(data[:, 1: nfreq].conjugate())


def amax_cc(sig):
    return np.argmax(sig) - len(sig) // 2


def angle(a, b):
    # return np.arctan((a[1] - b[1]) / (a[0] - b[0]))

    return np.arctan2((a[1] - b[1]), (a[0] - b[0]))


def add_noise(a, freqs, sr, scale, taplen=0.05):
    out = a.copy()

    nsig, npts = a.shape
    fwin = freq_window(freqs, npts, sr)
    nfreq = len(fwin)

    for i in range(nsig):
        fb = np.zeros(npts, dtype=np.complex64)

        phases = np.random.rand(nfreq) * 2 * np.pi
        phases = np.cos(phases) + 1j * np.sin(phases)

        fb[: nfreq] = phases * fwin
        fb[-nfreq + 1:] = fb[1:nfreq].conjugate()[::-1]
        # a[i] = np.real(ifft(fb))
        out[i] += np.real(ifft(fb)) * scale
    taper2d(out, int(taplen * npts))

    return out


def zeropad2d(a, npad):
    nrow, ncol = a.shape
    out = np.zeros((nrow, ncol + npad), dtype=a.dtype)
    out[:, :ncol] = a

    return out


def zero_argmax(a, wlen, taplen=0.05):
    # handles edges incorrectly
    npts = a.shape[1]
    taper = hann_half(int(taplen * wlen))
    win = np.concatenate((taper[::-1], np.zeros(wlen), taper))
    hl = int(len(win) // 2)
    out = a.copy()
    imaxes = np.argmax(np.abs(out), axis=1)

    for i, imax in enumerate(imaxes):
        i0 = imax - hl
        i1 = imax + hl

        if i0 <= 0:
            out[i, 0:i1] *= win[abs(i0):]
        elif i1 > npts:
            out[i, i0:] *= win[:-(i1 - npts)]
        else:
            out[i, i0:i1] *= win

    return out


def hann(npts):
    return 0.5 - 0.5 * np.cos(2.0 * np.pi * np.arange(npts) / (npts - 1))


def hann_half(npts):
    return hann(npts * 2)[:npts]


def fftnoise(f):
    f = np.array(f, dtype='complex')
    npts = (len(f) - 1) // 2
    phases = np.random.rand(npts) * 2 * np.pi
    phases = np.cos(phases) + 1j * np.sin(phases)
    f[1:npts + 1] *= phases
    f[-1:-1 - npts:-1] = np.conj(f[1:npts + 1])

    return ifft(f).real


def band_noise(band, sr, samples):
    freqmin, freqmax = band
    freqs = np.abs(fftfreq(samples, 1. / sr))
    f = np.zeros(samples)
    idx = np.where(np.logical_and(freqs >= freqmin, freqs <= freqmax))[0]
    f[idx] = 1

    return fftnoise(f)


def envelope(data):
    slen = len(data)
    FFT = fft(data, slen)
    FFT[1: slen // 2] *= 2
    FFT[slen // 2:] = 0

    return np.abs(ifft(FFT))


def read_csv(filename, site_code='', **kwargs):
    """
    read a csv file containing site information
    The first line of the csv file should contain the site name
    The expected file structure is as follows and should contain one header line
    <network>, <site name>, <site type>, <no component>, x, y, z
    where x, y and z represents the location of the sites expressed in a local
    coordinate system. Note that the <site name> is limited to four character
    because of NonLinLoc limitations.

    example of file strucuture

    1. <Network>, <site long name>, <site code>, <site type>, <gain>,
    <sensitivity>, <sx>, <sy>, <sz>, <channel 1 code>, <azimuth>, <dip>,
    <channel 2 code>, <azimuth>, <dip>, <channel 3 code>, <azimuth>, <dip>

    :param filename: path to a csv file
    :type filename: string
    :param site_code: site code
    :type site_code: string
    :param has_header: whether or not the input file has an header
    :type has_header: bool
    :rparam: site object
    :rtype: ~uquake.core.station.Site
    """

    from uquake.core.data.station import Site, Network, Station, Channel
    from numpy import loadtxt

    data = loadtxt(filename, delimiter=',', skiprows=1, dtype=object)
    stations = []

    for i, tmp in enumerate(data):

        nc, long_name, sc, st, smt, gain, sensitivity = tmp[:7]
        staloc = tmp[7:10].astype(float)
        orients = tmp[10:22].reshape(3, 4)

        channels = []

        for comp in orients:
            if not comp[0]:
                continue
            xyz = comp[1:4].astype(float)
            channel = Channel(code=comp[0])
            channel.orientation = xyz
            channels.append(channel)

        station = Station(long_name=long_name, code=sc, site_type=st,
                          motion_type=smt, gain=float(gain),
                          sensitivity=float(sensitivity), loc=staloc,
                          channels=channels)

        stations.append(station)

    networks = [Network(code=nc, stations=stations)]
    site = Site(code=site_code, networks=networks)

    return site



""" Collection of functions to calculate moment magnitude

"""

import warnings

import numpy as np
from uquake.helpers.logging import logger
from obspy.core.event import Comment, ResourceIdentifier, WaveformStreamID
from uquake.core.event import Magnitude
from obspy.core.event.magnitude import (StationMagnitude,
                                        StationMagnitudeContribution)

from uquake.core.event import Pick
from uquake.waveform.amp_measures import measure_pick_amps
from uquake.waveform.mag_utils import double_couple_rad_pat, \
    free_surface_displacement_amplification
from uquake.waveform.smom_measure_legacy import measure_pick_smom

warnings.simplefilter("ignore", UserWarning)
warnings.simplefilter("ignore")


def moment_magnitude_new(st, event,
                         vp=5300, vs=3500, density=2700,
                         only_triaxial=True,
                         min_dist=20,
                         fmin=20, fmax=1000,
                         use_smom=False):
    # TODO: this moment_magnitude_new function looks broken
    picks = event.picks

    if use_smom:
        measure_pick_smom(st, picks, debug=False)
        comment = "Average of freq-domain P & S moment magnitudes"
    else:
        measure_pick_amps(st, picks, debug=False)
        comment = "Average of time-domain P & S moment magnitudes"

    # Use time(or freq) lambda to calculate moment magnitudes for each arrival

    Mw_P, station_mags_P = calc_magnitudes_from_lambda(cat, vp=vp, vs=vs,
                                                       density=density,
                                                       P_or_S='P',
                                                       use_smom=use_smom)

    Mw_S, station_mags_S = calc_magnitudes_from_lambda(cat, vp=vp, vs=vs,
                                                       density=density,
                                                       P_or_S='S',
                                                       use_smom=use_smom)

    # Average Mw_P,Mw_S to get event Mw and wrap with list of
    #   station mags/contributions

    Mw = 0.5 * (Mw_P + Mw_S)

    station_mags = station_mags_P + station_mags_S
    set_new_event_mag(event, station_mags, Mw, comment)


def set_new_event_mag(event, station_mags, Mw, comment, make_preferred=False):
    count = len(station_mags)

    sta_mag_contributions = []

    for sta_mag in station_mags:
        sta_mag_contributions.append(StationMagnitudeContribution(
            station_magnitude_id=sta_mag.resource_id)
        )

    origin_id = event.preferred_origin().resource_id

    event_mag = Magnitude(origin_id=origin_id,
                          mag=Mw,
                          magnitude_type='Mw',
                          station_count=count,
                          evaluation_mode='automatic',
                          station_magnitude_contributions=sta_mag_contributions,
                          comments=[Comment(text=comment)],
                          )

    event.magnitudes.append(event_mag)
    event.station_magnitudes = station_mags

    if make_preferred:
        event.preferred_magnitude_id = ResourceIdentifier(
            id=event_mag.resource_id.id,
            referred_object=event_mag)

    return


def calc_magnitudes_from_lambda(cat,
                                vp=5300, vs=3500, density=2700,
                                P_or_S='P',
                                use_smom=False,
                                use_sdr_rad=False,
                                use_free_surface_correction=False,
                                min_dist=20.,
                                **kwargs):
    """
    Calculate the moment magnitude at each station from lambda,
      where lambda is either:
        'dis_pulse_area' (use_smom=False) - calculated by integrating arrival
            displacement pulse in time
        'smom' (use_smom=True) - calculated by fiting Brune spectrum to
            displacement spectrum in frequency
    """

    fname = 'calc_magnitudes_from_lambda'

    # Don't loop over event here, do it in the calling routine
    #   so that vp/vs can be set for correct source depth
    event = cat[0]
    origin = event.preferred_origin() if event.preferred_origin() else \
        event.origins[0]
    ev_loc = origin.loc
    origin_id = origin.resource_id

    rad_P, rad_S = 0.52, 0.63

    if P_or_S == 'P':
        v = vp
        rad = rad_P
        mag_type = 'Mw_P'
    else:
        v = vs
        rad = rad_S
        mag_type = 'Mw_S'

    if use_smom:
        magnitude_comment = 'station magnitude measured in frequeny-domain (' \
                            'smom)'
        lambda_key = 'smom'
    else:
        magnitude_comment = 'station magnitude measured in time-domain (' \
                            'dis_pulse_area)'
        lambda_key = 'dis_pulse_area'

    if use_free_surface_correction and np.abs(ev_loc[2]) > 0.:
        logger.warning("%s: Free surface correction requested for event ["
                       "h=%.1f] > 0" % (fname, ev_loc[2]))

    if use_sdr_rad and 'sdr' not in kwargs:
        logger.warning("%s: use_sdr_rad requested but NO [sdr] given!" % fname)

    station_mags = []
    Mw_list = []

    Mw_P = []

    arrivals = [arr for arr in event.preferred_origin().arrivals if
                arr.phase == P_or_S]

    for arr in arrivals:

        try:
            pk = arr.pick_id.get_referred_object()
            sta = pk.waveform_id.station_code
            cha = pk.waveform_id.channel_code
            net = pk.waveform_id.network_code
        except AttributeError:
            logger.warning('Missing data on arrival', exc_info=True)

            continue

        fs_factor = 1.

        if use_free_surface_correction:
            if arr.get('inc_angle', None):
                inc_angle = arr.inc_angle
                fs_factor = free_surface_displacement_amplification(
                    inc_angle, vp, vs, incident_wave=P_or_S)

                # MTH: Not ready to implement this.  The reflection coefficients
                #      are expressed in x1,x2,x3 coords
                # print("inc_angle:%.1f x1:%.1f x3:%.1f" % (inc_angle, fs_factor[0], fs_factor[2]))
                # MTH: The free surface corrections are returned as <x1,x2,x3>=<
                fs_factor = 1.
            else:
                logger.warning("%s: sta:%s cha:%s pha:%s: inc_angle NOT set "
                               "in arrival dict --> use default" %
                               (fname, sta, cha, arr.phase))

        if use_sdr_rad and 'sdr' in kwargs:
            strike, dip, rake = kwargs['sdr']

            if arr.get('takeoff_angle', None) and arr.get('azimuth', None):
                takeoff_angle = arr.takeoff_angle
                takeoff_azimuth = arr.azimuth
                rad = double_couple_rad_pat(takeoff_angle, takeoff_azimuth,
                                            strike, dip, rake, phase=P_or_S)
                rad = np.abs(rad)
                logger.debug("%s: phase=%s rad=%f" % (fname, P_or_S, rad))
                magnitude_comment += ' radiation pattern calculated for (s,' \
                                     'd,r)= (%.1f,%.1f,%.1f) theta:%.1f ' \
                                     'az:%.1f pha:%s |rad|=%f' % \
                                     (strike, dip, rake, takeoff_angle,
                                      takeoff_azimuth,
                                      P_or_S, rad)
                # logger.info(magnitude_comment)
            else:
                logger.warnng("%s: sta:%s cha:%s pha:%s: "
                              "takeoff_angle/azimuth NOT set in arrival dict --> use default radiation pattenr" %
                              (fname, sta, cha, arr.phase))

        _lambda = getattr(arr, lambda_key)

        if _lambda is not None:

            M0_scale = 4. * np.pi * density * v ** 3 / (rad * fs_factor)

            # R  = np.linalg.norm(sta_dict['station'].loc -ev_loc) #Dist in meters

            # MTH: obspy arrival.distance = *epicentral* distance in degrees
            #   >> Add attribute hypo_dist_in_m to uquake arrival class
            #         to make it clear
            if arr.distance:
                R = arr.distance
            else:
                R = arr.hypo_dist_in_m

            if R >= min_dist:

                M0 = M0_scale * R * np.abs(_lambda)
                Mw = 2. / 3. * np.log10(M0) - 6.033
                # print("MTH: _lambda=%g R=%.1f M0=%g" % (np.abs(_lambda), R, M0))

                Mw_list.append(Mw)

                station_mag = StationMagnitude(origin_id=origin_id, mag=Mw,
                                               station_magnitude_type=mag_type,
                                               comments=[Comment(
                                                   text=magnitude_comment)],
                                               waveform_id=WaveformStreamID(
                                                   network_code=net,
                                                   station_code=sta,
                                                   channel_code=cha))
                station_mags.append(station_mag)

            else:
                logger.info("arrival sta:%s pha:%s dist=%.2f < min_dist("
                            "=%.2f) --> Skip" % (fname, sta, arr.phase, R,
                                                 min_dist))

        # else:
        # logger.warning("arrival sta:%s cha:%s arr pha:%s lambda_key:%s is NOT SET --> Skip" \
        # % (sta, cha, arr.phase, lambda_key))

    logger.info("nmags=%d avg:%.1f med:%.1f std:%.1f" %
                (len(Mw_list), np.mean(Mw_list), np.median(Mw_list),
                 np.std(Mw_list)))

    return np.median(Mw_list), station_mags


def calculate_energy_from_flux(cat,
                               inventory,
                               vp,
                               vs,
                               rho=2700.,
                               use_sdr_rad=False,
                               use_water_level=False,
                               rad_min=0.2):
    fname = 'calculate_energy_from_flux'

    for event in cat:
        origin = event.preferred_origin() if event.preferred_origin() else \
        event.origins[0]

        use_sdr = False

        if use_sdr_rad and event.preferred_focal_mechanism() is not None:
            mech = event.preferred_focal_mechanism()
            np1 = event.preferred_focal_mechanism().nodal_planes.nodal_plane_1
            sdr = (np1.strike, np1.dip, np1.rake)
            use_sdr = True

        # for phase in ['P', 'S']:
        # for arr in [x for x in arrivals if x.phase == phase]:

        P_energy = []
        S_energy = []

        for arr in origin.arrivals:
            pk = Pick(arr.get_pick())
            # try:
            sta = pk.get_sta()
            sta_response = inventory.select(sta)
            # except AttributeError:
            #     logger.warning(
            #         f'Cannot get station for arrival "{arr.resource_id}"'
            #         f' for event "{event.resource_id}".')

            # continue
            phase = arr.phase

            if phase.upper() == 'P':
                velocity = vp.interpolate(sta_response.loc)
                rad_pat = 4 / 15

            elif phase.upper() == 'S':
                velocity = vs.interpolate(sta_response.loc)
                rad_pat = 2 / 5

            # could check for arr.hypo_dist_in_m here but it's almost identical
            R = arr.distance

            # MTH: Setting preferred flux = vel_flux_Q = attenuation tstar
            # corrected flux
            flux = 0

            if arr.vel_flux_Q is not None:
                flux = arr.vel_flux_Q
                logger.info("%s: vel_flux_Q exists in [%s], using this for "
                            "energy, arr for sta:%s" %
                            (fname, phase, sta))
            elif arr.vel_flux is not None:
                flux = arr.vel_flux
                logger.info("%s: vel_flux exists in [%s], using this for "
                            "energy, arr for sta:%s" %
                            (fname, phase, sta))
            else:
                logger.info("%s: No vel_flux set for arr sta:%s pha:%s --> "
                            "skip energy calc" %
                            (fname, sta, phase))

                continue

            energy = (4. * np.pi * R ** 2) * rho * velocity * flux

            scale = 1.

            if use_sdr:
                if arr.get('takeoff_angle', None) and arr.get('azimuth', None):
                    takeoff_angle = arr.takeoff_angle
                    takeoff_azimuth = arr.azimuth
                    strike = sdr[0]
                    dip = sdr[1]
                    rake = sdr[2]
                    rad = double_couple_rad_pat(takeoff_angle, takeoff_azimuth,
                                                strike, dip, rake, phase=phase)

                    if use_water_level and np.abs(rad) < rad_min:
                        rad = rad_min

                    scale = rad_pat / rad ** 2

            energy *= scale

            arr.energy = energy

            if phase == 'P':
                P_energy.append(energy)
            else:
                S_energy.append(energy)

        if not P_energy:
            P_energy = [0]
            logger.warning('No P energy measurements. The P-wave energy will '
                           'be set to 0, the total energy will not include '
                           'the P-wave energy. This will bias the total '
                           'energy value')
        if not S_energy:
            S_energy = [0]
            logger.warning(
                'No S energy measurements. The S-wave energy will be set to '
                '0, the total energy will not include the P-wave energy. '
                'This will bias the total energy value')

        energy_p = np.median(P_energy)
        energy_s = np.median(S_energy)
        E = energy_p + energy_s

        nvals = len(S_energy) + len(P_energy)
        comment = 'Energy [N-m] calculated from sum of median P + median S ' \
                  'energy'
        comment_ep = '"Ep":{}, "std_Ep":{}'.format(np.median(P_energy),
                                                   np.std(P_energy))
        comment_ep = '{' + comment_ep + '}'
        comment_es = '"Es":{}, "std_Es":{}'.format(np.median(S_energy),
                                                   np.std(S_energy))
        comment_es = '{' + comment_es + '}'

        if E > 0:
            # Note: this is not a "mag" (there is no log10).
            #       Just using magnitude class to hold it in quakeml

            energy_mag = Magnitude(origin_id=origin.resource_id,
                                   mag=E,
                                   magnitude_type='E',
                                   station_count=nvals,
                                   evaluation_mode='automatic',
                                   comments=[Comment(text=comment),
                                             Comment(text=comment_ep),
                                             Comment(text=comment_es)],
                                   )
            energy_mag.energy_p_joule = energy_p
            energy_mag.energy_s_joule = energy_s
            energy_mag.energy_joule = E

            event.magnitudes.append(energy_mag)

        else:
            logger.warning("%s: Calculated val of Energy E=[%s] nS=%d nP=%d "
                           "is not fit to keep!"
                           % (fname, E, nvals, len(P_energy)))
            logger.warning("%s: Energy mag not written to Quakeml" % fname)

    return cat


if __name__ == '__main__':
    main()



from io import BytesIO
from tqdm import tqdm
import requests


def download_file_from_url(url):
    response = requests.get(url, stream=True)

    total_size = int(response.headers.get('content-length', 0))
    block_size = 1024
    progress_bar = tqdm(total=total_size, unit='iB', unit_scale=True)

    binary_file_object = BytesIO()
    for data in response.iter_content(block_size):
        progress_bar.update(len(data))
        binary_file_object.write(data)

    progress_bar.close()

    # Set the position of the BytesIO object back to the beginning
    binary_file_object.seek(0)

    return binary_file_object



# from obspy.core.util.base import *
from subprocess import call

from obspy.core.util.base import ENTRY_POINTS, _get_entry_points

# appending elements to the obspy ENTRY_POINTS
ENTRY_POINTS['grid'] = _get_entry_points('uquake.io.grid', 'readFormat')
ENTRY_POINTS['grid_write'] = _get_entry_points('uquake.io.grid',
                                               'writeFormat')

gfr_entry_points = _get_entry_points('uquake.io.waveform', 'readFormat')
gfw_entry_points = _get_entry_points('uquake.io.waveform', 'writeFormat')

wf_entry_points = _get_entry_points('uquake.io.waveform', 'readFormat')

for key in wf_entry_points.keys():
    ENTRY_POINTS['waveform'][key] = wf_entry_points[key]

wfw_entry_points = _get_entry_points('uquake.io.waveform', 'writeFormat')

for key in wfw_entry_points.keys():
    ENTRY_POINTS['waveform_write'][key] = wfw_entry_points[key]

evt_entry_points = _get_entry_points('uquake.io.event', 'readFormat')

for key in evt_entry_points.keys():
    ENTRY_POINTS['event'][key] = evt_entry_points[key]

invr_entry_points = _get_entry_points('uquake.io.inventory', 'readFormat')
for key in invr_entry_points.keys():
    ENTRY_POINTS['inventory'][key] = invr_entry_points[key]


def proc(cmd, cwd='.', silent=True):
    from ..logging import logger

    try:
        if silent:
            cmd = '%s > /dev/null 2>&1' % cmd
        retcode = call(cmd, shell=True, cwd=cwd)

        if retcode < 0:
            logger.error('Child was terminated by signal %d' % (retcode,))
        # else:
        # print >>sys.stderr, "Child returned", retcode
    except OSError as e:
        logger.error('Execution failed: %s' % (e,))


def align_decimal(number, left_pad=7, precision=2):
    """Format a number in a way that will align decimal points."""
    outer = '{0:>%i}.{1:<%i}' % (left_pad, precision)
    inner = '{:.%if}' % (precision,)

    return outer.format(*(inner.format(number).split('.')))


def pretty_print_array(arr):
    return '(%s)' % ''.join([align_decimal(a) for a in arr])


def np_array(arr):
    new_arr = np.empty(shape=(len(arr),), dtype=object)

    for i, el in enumerate(arr):
        new_arr[i] = el
    plugin_
    return new_arr


def _read_from_plugin(type, filename, format=None, **kwargs):
    """
    Reads a single file from a plug-in's readFormat function.
    """
    eps = ENTRY_POINTS[plugin_type]
    # get format entry point
    format_ep = None

    if not format:
        # auto detect format - go through all known formats in given sort order

        for format_ep in eps.values():
            # search isFormat for given entry point
            is_format = load_entry_point(
                format_ep.dist.key,
                'obspy.plugin.%s.%s' % (plugin_type, format_ep.name),
                'isFormat')
            # If it is a file-like object, store the position and restore it
            # later to avoid that the isFormat() functions move the file
            # pointer.

            if hasattr(filename, "tell") and hasattr(filename, "seek"):
                position = filename.tell()
            else:
                position = None
            # check format
            is_format = is_format(filename)

            if position is not None:
                filename.seek(0, 0)

            if is_format:
                break
        else:
            raise TypeError('Unknown format for file %s' % filename)
    else:
        # format given via argument
        format = format.upper()
        try:
            format_ep = eps[format]
        except (KeyError, IndexError):
            msg = "Format \"%s\" is not supported. Supported types: %s"
            raise TypeError(msg % (format, ', '.join(eps)))
    # file format should be known by now
    try:
        # search readFormat for given entry point
        read_format = load_entry_point(
            format_ep.dist.key,
            'obspy.plugin.%s.%s' % (plugin_type, format_ep.name),
            'readFormat')
    except ImportError:
        msg = "Format \"%s\" is not supported. Supported types: %s"
        raise TypeError(msg % (format_ep.name, ', '.join(eps)))
    # read
    list_obj = read_format(filename, **kwargs)

    return list_obj, format_ep.name



import numpy as np
from numpy.fft import fft
from scipy.optimize import curve_fit
from scipy.ndimage.interpolation import map_coordinates
from uquake.core.logging import logger
from obspy.core.trace import Stats
from uquake.core.stream import Trace, Stream
from uquake.grid.base import Grid
from uquake.core import event


########################################################################################
# Function defining attenuation relationship
########################################################################################


def geometrical_spreading(raypath):
    segment_len = np.sqrt(np.sum(np.diff(raypath, axis=0) ** 2, axis=1))

    # Source receiver distance
    r = np.sum(segment_len)

    if r < 1:
        r = 1

    # Geometrical spreading attenuation
    Ar = 1 / r

    return Ar


def anelastic_scattering_attenuation(raypath_or_distance, velocity, quality,
                                     seismogram, return_seimogram=True):
    """
    Calculate the anelastic_scattering_attenuation
    :param raypath_or_distance: raypath or distance
    :type raypath_or_distance: a list of point along the raypath or a float
    :param velocity: velocity along the raypath
    :type: uquake.core.data.grid.Grid or float
    :param quality: Seismic quality factor
    :type quality: uquake.core.data.grid.Grid or float
    :param seismogram: displacement waveform
    :type seismogram: uquake.core.Trace
    :return: attenuated seismogram
    :rtype: uquake.core.Trace
    """

    freq_coeff, _ = interpolate_Fc_Mw()
    freq_corner = FcMw(Mw, freq_coeff[0], freq_coeff[1])

    f = np.arange(0, 1000) / 1000.0 * freq_corner  # calculating the
    # attenuation up to the corner frequency
    Aq = np.exp(- np.pi * f)

    ray = quality.transform_to(raypath)

    Qray = map_coordinates(quality.data, ray.T, mode='nearest')
    Vray = map_coordinates(velocity.data, ray.T, mode='nearest')
    slen = np.sqrt(np.sum(np.diff(raypath, axis=0) ** 2, axis=1))

    Att = np.zeros(f.shape)
    for k, segment in enumerate(ray[:-1, :]):
        Qval = np.mean(Qray[k:k + 2])
        Vval = np.mean(Vray[k:k + 2])
        Att += slen[k] / (Vval * Qval)

    Aq_f = np.exp(-np.pi * f * Att)
    Aq = np.mean(Aq_f)

    return Aq


def radiation_pattern_attenuation():
    """
    :rparam: radiation pattern attenuation for S- and S-waves
    :rtype: tuple of float

    .. NOTE::

    Returned values are taken from Aki and Richard (2002) which are
    0.52 and 0.63 for P- and S-waves, respectively.

    .. EXAMPLE::

    >>> (PRadiation, SRadiation) = radiation_pattern_attenuation()
    """
    return 0.52, 0.63


def geometrical_spreading_attenuation(raypath, velocity=None, quality=None,
                                      Mw=-1):
    """
    :param raypath: Raypath between source and receiver
    :type raypath: nympy array
    :param velocity: velocity grid (default None)
    :type velocity: uquake ImageData
    :param quality: quality factor (default None)
    :type quality: uquake ImageData
    :return: the a tuple containing the geometrical spreading attenuation
    and the attenuation related to anelastic absorbtion and scattering
    """

    ar = geometrical_spreading(raypath)
    if quality:
        aq = anelastic_scattering_attenuation(raypath, velocity, quality, Mw)
        att = ar * aq
    else:
        att = ar

    return att


def calculate_attenuation(raypath, velocity, quality=None, Mw=-1):
    """
    :param source: source location tuple dimension of the grid
    :param receiver: receiver location tuple dimension of the grid
    :param velocity: velocity in EKImageData format
    :param quality: quality factor grid in EKImageData format
    :return the a tuple containing the geometrical spreading attenuation and
    the attenuation related to anelastic absorbtion and scattering
    """

    ar = geometrical_spreading(raypath)
    if quality:
        aq = anelastic_scattering_attenuation(raypath, velocity, quality, Mw)
        att = ar * aq
    else:
        att = ar

    return att


def calculate_attenuation_grid(seed, velocity, quality=None, locations=None,
                               triaxial=True,
                               orientation=(0, 0, 1), pwave=True, buf=0,
                               traveltime=None, eventSeed=False, Mw=-1.,
                               tt=None, return_tt=False,
                               homogeneous=True):
    """
    :param seed: seed (often receiver) location
    :type seed: tuple with same dimension as the grid
    :param velocity: velocity grid
    :type velocity: uquake.core.data.Grid
    :param locations: 2-D grid containing the coordinates at which the
    attenuation is calculated
    for instance [[x1,y1,z1],[x2,y2,z2],...,[xn,yn,zn]].
    :type location: 2D numpy array.
    :param quality: quality factor grid
    :type quality: uquake ImageData
    :param triaxial: true if triaxial sensor is used
    :type triaxial: bool
    :param orientation: sensor orientation taken into account only for
    uniaxial (non triaxial) sensors
    :type orientation: tuple
    :param pwave: true for p-wave
    :type pwave: bool
    :param progress: show progress bar if true
    :type progress: bool
    :param traveltime: precalculated traveltime grid (for MapReduce)
    :type traveltime: uquake.core.data.Grid
    :param eventSeed: True if seed is an event
    :type eventSeed: bool
    :param Mw: Moment magnitude
    :type Mw: float
    :param return_tt: if True, return traveltime grid
    :type return_tt: bool
    :param homogeneous: if True grid is considered homogeneous. Only
    geometrical spreading and radiation patterns
    are accounted for
    :type homogeneous: bool
    :rparam: Attenuation on a grid
    :rtype: uquake ImageData and

    .. note::

    The attenuation is calculated only at points within the velocity grid.
    Attenuation outside of the velocity grid have a value of 0.

    To generate the location grid, one could use the following:
        location = [coord for coord in zip(X, Y, Z)], where X, Y, and Z are
        vectors

    The script works both in 2D and 3D.

    .. example::

    """

    from uquake.simul import eik

    A = []

    if not np.any(locations):
        providedInputLocation = False
        x = np.arange(0, velocity.shape[0]) * velocity.spacing + \
            velocity.origin[0]
        y = np.arange(0, velocity.shape[1]) * velocity.spacing + \
            velocity.origin[1]
        z = np.arange(0, velocity.shape[2]) * velocity.spacing + \
            velocity.origin[2]

        X, Y, Z = np.meshgrid(x, y, z)

        locations = [coord for coord in zip(X.ravel(), Y.ravel(), Z.ravel())]
    else:
        providedInputLocation = True

    NoNode = len(locations)

    c1 = velocity.origin
    c2 = velocity.origin + (np.array(velocity.shape) - \
                            np.array([1, 1, 1])) * velocity.spacing

    for k, coord in enumerate(locations):
        pb()

        # if (np.any(coord - c1 < buf)) or (np.any(c2 - coord < buf)):
        #   A.append(0)
        #   continue

        dist = np.linalg.norm(coord - seed)

        if homogeneous:
            ray = np.array([seed, coord])
            quality = None
        else:
            if not tt:
                tt = eik.EikonalSolver(velocity, seed)

            ray = eik.RayTracer(tt, velocity, np.array(coord))

        if eventSeed:
            ray = ray[-1::-1, :]

        att = calculate_attenuation(ray, velocity, quality, Mw=Mw)

        att_comp = 1
        if not triaxial:

            incidenceVector = tmp / np.linalg.norm(tmp)
            tmpatt = np.dot(incidenceVector, np.array(orientation))
            if pwave:
                att_comp = tmpatt
            else:
                att_comp = np.sqrt(1 - tmpatt ** 2)

        A.append(att * att_comp)

    if not providedInputLocation:
        tmp = np.array(A).reshape(velocity.shape)
    else:
        tmp = np.array(A)

    if return_tt:
        if not tt:
            tt = eik.EikonalSolver(velocity, seed)

        import scipy.ndimage as ndimage

        tt_out = Grid(
            ndimage.map_coordinates(tt.data, tt.transform_to(locations).T),
            spacing=velocity.shape, origin=velocity.origin)
        return Grid(tmp, spacing=velocity.shape,
                        origin=velocity.origin), tt_out
    else:
        return Grid(tmp, spacing=velocity.shape, origin=velocity.origin)


def Mw2M0(Mw):
    """Converts Moment Magnitude, Mw, into Seismic Moment, M0.
    The relation between Moment magnitude and seismic moment is
        Mw = 2/3 log_10(M0) - 6.02 (Hank and Kanamori, 1979)

    :param Mw: Moment magnitude
    :type Mw: float
    :returns: Seismic moment
    :rtype: float
    """

    return 10 ** ((3 / 2.0) * (Mw + 6.07))


def M02Mw(M0):
    """Converts Seismic Moment, M0 into Moment Magnitude, Mw,.
    The relation between Moment magnitude and seismic moment is
        Mw = 2/3 log_10(M0) - 6.02 (Hank and Kanamori, 1979)

    :param Mw: Moment magnitude
    :type Mw: float
    :returns: Seismic moment
    :rtype: float
    """

    return 2 / 3.0 * np.log10(M0) - 6.07


def corner_frequency(Mw, vp=5000.0, vs=3500.0, SSD=1):
    """Calculates the corner frequency, f0, from the moment magnitude,
    Mw, we use the following relationship

    log10(f0) = 1.32 + 0.33 * log10(SSD) + log10(v) - 0.5 * Mw,

    where SSD is the static stress drop in "bar", and v the velocity in "km/s".

    --> Carefully note the unit of the input parameters

    :param Mw: Moment magnitude of the event
    :type Mw: float
    :param vp: p-wave velocity in m/s
    :type vp: float
    :param vs: s-wave velocity in m/s
    :type vs: float
    :param SSD: Static stress drop in "bar" generally though to change
    from 1 - 100 bar.
    :type SSD: float
    :returns: f0_p and f0_s, the corner frequency in Hz for the P- and S-waves,
    respectively (f0_p and f0_s).
    :rtype: tuple of float

    .. note:: For tectonic earthquake, the SSD is reported to be between
    1 - 100.
    For microseismic event, the SSD is generally close to 1 and can be lower.
    """

    f0_p = 10 ** (
            1.32 + 0.33 * np.log10(SSD) + np.log10(vp / 1000.0) - 0.5 * Mw)
    f0_s = 10 ** (
            1.32 + 0.33 * np.log10(SSD) + np.log10(vs / 1000.0) - 0.5 * Mw)

    return (f0_p, f0_s)


def synthetic_seismogram(Mw, duration=0.1, sampling_rate=10000, vp=5000.0,
                         vs=3500.0, rho=2400, SSD=1, pwave=True):
    """
    Create a synthetic displacement pulse at the source seismogram based
    on the brune model (Brune 1970).
    This model is extensively used and generally agrees with observations from
    many different setting and over a large range of magnitude.

    The displacement time function, u(t), is expressed as follows:

        u(t) = A_0 x t x omega_0 x H(t) * exp(-t x omega_0) ,

    where t the time, omega_0, the angular frequency and H(t) is the
    heavyside function. Note that the angular frequency is calculated from
    the static stress drop (SSD). A0 is given by the following
    equation:

        A0 = M0 / (4 * pi * rho * v ** 3)

    References for further reading:
    - Routine data processing in earthquake seismology
    - Relating Peak Particle Velocity and Acceleration to Moment Magnitude
    in Passive (Micro-) Seismic Monitoring
    (www.bcengineers.com/images/BCE_Technical_Note_3.pdf)

    :param Mw: the moment magnitude of the seismic event (default: -1),
    this value determine the wave amplitude and the
    frequency content
    :type Mw: float
    :param noise_level: level of gaussian noise to add to the synthetic
    seismogram (default: 1e-5)
    :type noise_level: float
    :param duration: duration of the seismogram in seconds (default: 0.1),
    the pulse is centered at zero
    :type duration: float
    :param sampling_rate: sampling rate in Hz of the generated time series (
    default: 10000)
    :type sampling_rate: int
    :param vp: P-wave velocity of the material at the source (default: 5000 m/s)
    :type vp: float
    :param vs: S-wave velocity of the material at the source (default: 3500 m/s)
    :type vs: float
    :param rho: density of the material at the source in kg/m**3 (default:
    2400 kg/m**3)
    :param SSD: Static stress drop in "bar" (default: 1 bar)
    :type SSD: float
    :param pwave: Return P-wave displacement seismogram if True and S-wave
    displacement seismogram if false
    :rparam: tuple Obspy Trace containing the seismogram
    :rtype: Obspy Trace

    .. note::
        The velocity and acceleration can easily be obtained by
        differentiating the trace using the Obspy Trace method
        differentiate.

        Example
        >>> tr = synthetic_seismogram(-1)
        >>> tr.differentiate()  # this creates a velocity trace
        >>> tr.differentiate()  # this creates an acceleration trace

        This operation is performed in place on the actual data arrays. The
        raw data is not accessible anymore afterwards. To keep your
        original data, use :meth:`~obspy.core.trace.Trace.copy` to create
        a copy of your trace object.
        This also makes an entry with information on the applied processing
        in ``stats.processing`` of this trace.

    """

    M0 = Mw2M0(Mw)
    (f0p, f0s) = corner_frequency(Mw, vp, vs, SSD)
    # duration = 5 / f0p
    npts = duration * sampling_rate
    t = np.arange(
        npts) / sampling_rate
    if pwave:
        W0 = 2 * np.pi * f0p
        v = vp
    else:
        W0 = 2 * np.pi * f0s
        v = vs

    A0 = M0 / (4 * np.pi * rho * v ** 3)

    data = A0 * t * W0 ** 2 * np.exp(-t * W0)
    data = np.roll(data, len(data) / 2)

    stats = Stats()
    stats.sampling_rate = sampling_rate
    stats.npts = 2 * npts - 1

    from uquake.core.util.cepstrum import minimum_phase
    minphase_data = np.roll(minimum_phase(data, len(data)), len(data) / 2)

    return Trace(data=data, header=stats)


def detection_level_sta_lta_grid(attenuationGrid, VpGrid, VsGrid,
                                 noise_level=1e-3, acceleration=True,
                                 STALTA_threshold=3, SSD=0.1,
                                 minMag=-3., maxMag=2., magResolution=0.1,
                                 pwave=True):
    """
    Returns a grid containing the minimum magnitude detectable at the
    location stloc
    with the same dimensions as the attenuation grid.

    :param attenuationGrid: Attenuation grid
    :type attenuationGrid: uquake.core.data.Grid
    :param VpGrid: P-wave velocity grid
    :type VpGrid: uquake.core.data.Grid
    :param VsGrid: S-wave velocity grid
    :type VsGrid: uquake.core.data.Grid
    :param noise_level: Level of noise to add to the seismogram
    :type noise_level: float
    :param acceleration: if True use acceleration if false use velocity
    :type acceleration: bool
    :param STALTA_threshold: STA/LTA threshold to declare a trigger.
    (default 3).
    :type STALTA_threshold: float
    :param SSD: Static stress drop in "bar" (default: 1 bar)
    :type SSD: float
    :param minMag: minimum moment magnitude to be tested (default=-3)
    :type minMag: float
    :param maxMag: maximum moment magnitude to be tested (default=2)
    :type maxMag: float
    :param magResolution: resolution of magnitude increments between minMag
    and maxMag
    :type magResolution: float
    :param pwave: True if it is a P-wave
    :type pwave: bool
    :rtype: uquake.core.data.Grid
    """

    Sensitivity = attenuationGrid.copy()
    Sensitivity.data[:] = maxMag
    magnitudes = np.arange(minMag, maxMag + magResolution, magResolution)
    for Mw in magnitudes[-1::-1]:
        Pulse = synthetic_seismogram(Mw, duration=1.5, sampling_rate=10000,
                                     vp=5000.0, vs=3500.0,
                                     rho=2400, SSD=SSD,
                                     pwave=pwave)  # Displacement

        Pulse.differentiate()  # Velocity
        if acceleration:
            Pulse.differentiate()  # Acceleration

        amp = np.std(Pulse.data)

        attAmp = attenuationGrid.data * amp
        attEnergy = attAmp ** 2

        ENoise = noise_level ** 2

        STALTA = attEnergy / ENoise

        Sensitivity.data[STALTA >= STALTA_threshold] = Mw

    return Sensitivity


def triggered_sensor_sta_lta_grid(stloc, Rho, Vp, Vs, Qp, Qs, Mw=-1,
                                  STALTA_threshold=3.0, noise_level=1e-5,
                                  acceleration=True, SSD=1, evloc=None,
                                  return_bool=True):
    """Return a boolean grid with same dimension as an input attenuation grid
    describing whether a sensor triggers or not.

    :param stloc: Location of the station to use to calculate the attenuation
    :type stloc: tuple of int or float
    :param Rho: Density grid (kg/m**3).
    :type Rho: ImageData
    :param Vp: P-wave velocity grid (m/s).
    :type Vp: ImageData
    :param Vs: S-wave velocity grid (m/s).
    :type Vs: ImageData
    :param Qp: Quality factor grid for P-wave.
    :type Qp: ImageData
    :param Qs: Quality factor grid for the S-wave.
    :type Qs: ImageData
    :param Mw: Moment magnitude of the event
    :type Mw: float
    :param STALTA_threshold: STA/LTA threshold to declare a trigger.
    (default 3).
    :type STALTA_threshold: float
    :param noise_level: Level of noise to add to the seismogram
    :type noise_level: float
    :param acceleration: if True use acceleration if false use velocity
    :type acceleration: bool
    :param evloc: Event location grid (optional). if None, simulation will
    be performed
    on each nodes of the property grids (Rho, Vp, Vs ...).
    :type evloc: ImageData
    :param return_bool: Return a boolean if True and int if False
    :type return_bool: bool
    :rparam: grid with the same dimension as the input grid
    :rtype: Numpy array of bool or int
    describing sensor triggering

    .. NOTE::

    The dimensions of Rho, Vp and Vs grids need to be the same.

    .. EXAMPLE::

    >>> from numpy.random import randn
    >>> from scipy.ndimage.filters import gaussian_filter
    >>> from uquake.data import ImageData
    >>> GridSpc = 10
    >>> Rho = ImageData(gaussian_filter(randn(100,100) * 100 + 2400.0, 5),
    spacing=GridSpc)
    >>> Vp = ImageData(gaussian_filter(randn(100,100) * 100 + 5000.0),
    spacing=GridSpc)
    >>> Vs = ImageData(gaussian_filter(randn(100,100) * 100 + 3500.0),
    spacing=GridSpc)
    >>> stloc = (100, 100)
    >>> Trigger = TriggerSensorSTALTAGrid(stloc, Rho, Vp, Vs)

    """

    # in the future that might be interesting to enable processing on EMR
    # for portion of codes
    # without having to explicitely use MR_JOB.
    # e.g.
    # 1) open EMR connection and create instances (minimal amount of bootstrap
    #    the executable and library needed should be dowloaded to the server
    #    using a git clone
    #    command for instance)
    # 2) send command to EMR culuster
    # 3) retreive information
    # 4) shutdown EMR cluster

    # need to be fixed...
    # if not evloc:
    #   pass
    if evloc.ndim < 2:
        evloc = evloc[np.newaxis, :]

    # Calculating attenuation grids
    (PRadiation, SRadiation) = radiation_pattern_attenuation()

    logger.info('Calculating P-wave attenuation')
    Att_p = PRadiation * calculate_attenuation_grid(stloc, Vp, evloc,
                                                    quality=Qp)
    logger.info('Calculating S-wave attenuation')
    Att_s = SRadiation * calculate_attenuation_grid(stloc, Vs, evloc,
                                                    quality=Qs)

    # need to better writer the function to pass arguments to this function
    # nargs*
    PulseP = synthetic_seismogram(Mw, duration=0.1, sampling_rate=10000,
                                  vp=5000.0, vs=3500.0,
                                  rho=2400, SSD=SSD,
                                  pwave=True)  # Displacement
    PulseS = synthetic_seismogram(Mw, duration=0.1, sampling_rate=10000,
                                  vp=5000.0, vs=3500.0,
                                  rho=2400, SSD=SSD, pwave=False)

    PulseP.differentiate()  # Velocity
    PulseS.differentiate()
    if acceleration:
        PulseP.differentiate()  # Acceleration
        PulseS.differentiate()  # Acceleration

    ESignalP = np.var(PulseP.data)
    ESignalS = np.var(PulseS.data)

    # Something wrong with last few lines here
    ESignalP_Grid = ESignalP * Att_p
    ESignalS_Grid = ESignalS * Att_s

    ESignal_Grid = ESignalP_Grid.copy()

    ESignal_Grid[ESignalP_Grid < ESignalS_Grid] = ESignalS_Grid[
        ESignalP_Grid < ESignalS_Grid]

    ENoise = noise_level ** 2

    STALTA = ESignal_Grid / ENoise
    Trigger = STALTA >= STALTA_threshold

    diff = STALTA - STALTA_threshold
    # print diff
    logger.info('STALTA DIFF: %s' % (diff))
    # logger.info('STALTA: %.5f  THRESH: %.5f   DIFF: %.5f' % (STALTA,
    # STALTA_threshold,STALTA - STALTA_threshold))
    logger.info('TRIGGERED?: %s' % (Trigger))

    if not return_bool:
        Trigger = Trigger.astype(int)

    return Trigger


# Beyond this point the code needs some (significant) clean up.
# ---------------------------------------------------

# defining the spectral function
def spectral_function(f, log_omega0, fc, q):
    """Defines the far-field Brune displacement spectrum
    (Brune, 1970).

    This function is defined in base-10 log for analytical purposes,
    as singularities can cause the function to fail.

    :param f: frequency at which to calculate the power spectrum
    :type f: float
    :param log_omega0: the base-10 log of the low-frequency level
    :type log_omega0: float
    :param fc: the corner frequency
    :type fc: float
    :param q: quality factor

    :returns: the base-10 log of the power spectrum associated
    with the given frequency f
    :rtype: float
    """
    # if np.any((1.0+(f/float(f0))**2) <= 0):

    # scattering attenuation
    # it is assumed that the correction for the travel time has been done
    # outside of this function.
    scat_att = np.exp(-np.pi * f / q)

    return log_omega0 + np.log10(scat_att) - np.log10((1.0 + (f / fc) ** 2))


def MwFc(fn, m, b):
    """Defines the moment magnitude for a given corner frequency

    The common use is with the help of scipy.optimize.curve_fit,
    to find the parameters m and b that fit a straight line in
    the frequency domain (see interpolate_Fc_Mw())

    :param fn: the corner frequency
    :type fn: float
    :param m: the slope of the linear fit
    :type m: float
    :param b: the intercept of the linear fit
    :type b: float

    :returns: the moment magnitude
    :rtype: float
    """
    return m * np.log10(fn) + b


def FcMw(Mw, m, b):
    """Defines the corner frequency for a given moment magnitude

    The common use is with the help of scipy.optimize.curve_fit,
    to find the parameters m and b that fit a straight line in
    the frequency domain (see interpolate_Fc_Mw())

    :param Mw: the moment magnitude
    :type Mw: float
    :param m: the slope of the linear fit
    :type m: float
    :param b: the intercept of the linear fit
    :type b: float

    :returns: the corner frequency
    :rtype: float
    """
    return 10 ** ((Mw - b) / m)


# we may want to modify this function to account for the source radium
def interpolate_Fc_Mw():
    """Linear regression of the moment magnitude - corner frequency
    empirical relationship

    This function uses scipy.optimize.curve_fit,
    to find the parameters m and b that fit a straight line in
    for the MwFc() function.

    :returns: the linear regression parameters m and b and their covariances
    :rtype: float
    """
    fc = [2000, 1000, 500, 100, 50, 25, 10]
    Mw = [-3, -2, -1, 0, 1, 2, 3]
    popt, pcov = curve_fit(MwFc, fc, Mw)

    return (popt, pcov)


# def Displacementfn(Mw, fc, Rho, V, Fc=0.52):
#   """Calculates the Displacement at the source
#   for a given homogeneous density and velocity model.
#   based on http://www.bcengineers.com/images/BCE_Technical_Note_3.pdf

#   The function calculates the displacement based on the following
#   relationship

#   D = M_0 x t x Omega_0 x exp()

#   :param Mw: the moment magnitude
#   :type Mw: float
#   :param fc: the corner frequency
#   :type fc: float
#   :param Rho: Density at the source
#   :type Rho: float
#   :param V: Seismic velocity at the source
#   :type V: float
#   :param Fc: radiation factor (commonly set to 0.52 or 0.63 for P and S
#   wave respectively [Aki and Richard, 2002])
#   :type Fc: float

#   """


#   # u(t) = M_0 x t x Omega_0 x H(t) x exp(-t x Omega_0)
#   pass

#   M0 = 10 ** ((Mw + 6.01) * (3 / 2))  # the seismic moment
#   w0 = 2 * np.pi * fc  # the corner angular frequency
#   PPV = M0 * Fc * (w0 ** 2) / (4 * np.pi * Rho * (V ** 3))
#   return PPV, w0

def PPVfn(Mw, fc, Rho, V):
    """Calculates the peak-particle-velocity (PPV) at the source
    for a given homogeneous density and velocity model.

    :param Mw: the moment magnitude
    :type Mw: float
    :param fc: the corner frequency in Hz
    :type fc: float
    :param Rho: Density at the source in kg/m**3
    :type Rho: float
    :param V: Seismic velocity at the source in m/s
    :type V: float

    :returns: the PPV
    :rtype: float
    """
    M0 = Mw2M0(Mw)  # the seismic moment
    w0 = 2 * np.pi * fc
    PPV = w0 ** 2 * M0 / (4 * np.pi * Rho * (V ** 3))
    return PPV


def PPAfn(Mw, fc, Rho, V):
    """Calculates the peak-particle-acceleration (PPA) at the source
    for a given homogeneous density and velocity model.

    :param Mw: the moment magnitude
    :type Mw: float
    :param fc: the corner frequency in Hz
    :type fc: float
    :param Rho: Density at the source in kg/m**2
    :type Rho: float
    :param V: Seismic velocity at the source in m/s
    :type V: float


    :returns: the PPA
    :rtype: float
    """
    PPV = PPVfn(Mw, fc, Rho, V)
    w0 = 2 * np.pi * fc
    PPA = np.abs(PPV * (1 - 2 * w0))
    return PPA


def Attenuation(fc, V, R, Q=100):
    """Defines the attenuation of a wave due to
    geometrical spreading and energy (quality factor)
    dissipation.

    :param fc: the corner frequency
    :type fc: float
    :param V: the homogeneous model velocity
    :type V: float
    :param Q: the homogeneous quality factor
    :type Q: float
    :param R: the distance
    :type R: float

    :returns: the attenuation factor
    :rtype: float
    """
    att = (1 / R) * np.exp(-np.pi * fc * R / (V * Q))
    return att


# see http://www.bcengineers.com/images/BCE_Technical_Note_3.pdf
# PPVThreshold=4e-5
def return_triggered_sensor(epos, spos, sorient=None, stype=None, Magnitude=-1,
                            V=5000, Rho=2400, Fc=0.6,
                            use_sensor_orientation=False,
                            use_sensor_type=False, PPAThreshold=0.02):
    # The correction for sensor orientation is the projection of the
    # incoming wavefront with respect to the sensor orientation
    # the result is a fraction e.g., 0.2

    # The correction for sensor type is related to the frequency response
    # the result will also be a fraction e.g., 0.5

    popt, pcov = interpolate_Fc_Mw()
    fc = FcMw(Magnitude, popt[0], popt[1])
    Mw = MwFc(fc, popt[0], popt[1])

    R = np.linalg.norm(spos - epos, axis=1)
    PPA = PPAfn(Mw, fc, Rho, V) * Attenuation(fc, V, R) * \
          radiation_pattern_attenuation()[0]

    indices = np.nonzero(PPA >= PPAThreshold)[0]
    return indices


class Trigger():

    def __init__(self, Mw, vp=5000.0, vs=3500.0, rho=2400, STALTA_threshold=3.,
                 SSD=0.02,
                 noise_level=0.02, acceleration=True):

        st = synthetic_seismogram(Mw, vp=vp, vs=vs, rho=rho, SSD=SSD,
                                  pwave=True)
        if acceleration:
            st.differentiate().differentiate()
        else:
            st.differentiate()

        self.amp = np.std(st.data)
        self.STALTA_threshold = STALTA_threshold
        self.noise_level = noise_level

    def Trigger(self, raypath):
        att = geometrical_spreading(raypath) * radiation_pattern_attenuation()[
            0]

        signal_energy = (att * self.amp) ** 2

        return signal_energy / (self.noise_level ** 2) > self.STALTA_threshold


class Sensitivity():

    def __init__(self, Mw_min=-3., Mw_max=3., Mw_spacing=0.1, vp=5000.0,
                 vs=3500.0,
                 rho=2400, STALTA_threshold=3., SSD=1,
                 noise_level=0.02, acceleration=True, pwave=True):

        nMw = (Mw_max - Mw_min) / Mw_spacing + 1
        self.Mw = np.linspace(Mw_min, Mw_max, nMw)
        self.pwave = pwave

        signal_amplitude = []

        for Mw_ in self.Mw:
            st = synthetic_seismogram(Mw_, vp=vp, vs=vs, rho=rho, SSD=SSD,
                                      pwave=pwave, duration=0.5)

            if acceleration:
                amp = np.std(st.differentiate().differentiate().data)
            else:
                amp = np.std(st.differentiate())

            amp = np.std(st.data)
            signal_amplitude.append(amp)

        self.signal_amplitude = np.array(signal_amplitude)
        self.STALTA_threshold = STALTA_threshold
        self.noise_level = noise_level
        self.Mw_max = Mw_max

    def Sensitivity(self, raypath):
        if self.pwave:
            att = geometrical_spreading(raypath) * \
                  radiation_pattern_attenuation()[0]
        else:
            att = geometrical_spreading(raypath) * \
                  radiation_pattern_attenuation()[1]

        signal_energy = (self.signal_amplitude * att) ** 2
        if np.any(np.abs(signal_energy / (
                self.noise_level ** 2)) > self.STALTA_threshold):
            return self.Mw[np.abs(signal_energy / (
                    self.noise_level ** 2)) > self.STALTA_threshold][0]
        else:
            return self.Mw_max


def trigger_sensor(raypath, Mw, vp=5000.0, vs=3500.0,
                   rho=2400, STALTA_threshold=3., SSD=0.02,
                   noise_level=0.02, acceleration=True):
    """
    Calculate the sensitivity at a specific location given a series sensor
    locations (location of senosor within an array)
    :param raypath: coordinates of the segment endpoints along the raypath
    :type raypath: numpy array
    :param Mw: Moment magnitude of the event
    :type Mw: float
    :param sensor_locs: sensor locations
    :type sensor_locs: numpy array
    :param Mw: Minimum moment magnitude (Mw) to consider (default -3)
    :type Mw_min: float
    :param Mw_max: Maximum moment magnitude (Mw) to consider (default 1)
    :type Mw_max: float
    :param Mw_spacing: resolution of the moment magnitude search (default 0.1)
    :type Mw_spacing: float
    :param vp: P-wave velocity at the source location in m/s (default 5000.)
    :type vp: float
    :param vs: S-wave velocity at the source location in m/s (default 3500.)
    :type vs: float
    :param rho: Density at the source location in kg/m**3 (default 2400.)
    :type rho: float
    :param STALTA_threshold: STA/LTA threshold for a trigger to be declared
    (default 3)
    :type STALTA_threshold: float
    :param SSD: Static stress drop (default 0.01)
    :type SSD: float
    :param noise_level: Noise amplitude level at sensor in the unit of
    sensor response (e.g. velocity or acceleration) (default 0.02 (m/s**2))
    :type noise_level: either float or numpy array with the same dimensions
    as sensor_loc
    :param acceleration: If true noise_level is assumed to be acceleration
    and calculation are performed consequently
    :type acceleration: bool
    :rparam: sensor trigger station
    :rtype: bool
    """

    att = geometrical_spreading(raypath) * radiation_pattern_attenuation()[0]
    st = synthetic_seismogram(Mw, vp=vp, vs=vs, rho=rho, SSD=SSD, pwave=True)
    if acceleration:
        st.differentiate().differentiate()
    else:
        st.differentiate()

    amp = np.std(att * st.data)

    return (amp ** 2) / (noise_level ** 2) > STALTA_threshold


def measure_sensitivity(raypath, Mw_min=-3., Mw_max=2., Mw_spacing=0.1,
                        vp=5000.0, vs=3500.0,
                        rho=2400, STALTA_threshold=3., SSD=0.02,
                        noise_level=0.02, acceleration=True):
    """
    Calculate the sensitivity at a specific location given a series sensor
    locations (location of senosor within an array)
    :param raypath: coordinates of the segment endpoints along the raypath
    :type raypath: numpy array
    :param sensor_locs: sensor locations
    :type sensor_locs: numpy array
    :param Mw_min: Minimum moment magnitude (Mw) to consider (default -3)
    :type Mw_min: float
    :param Mw_max: Maximum moment magnitude (Mw) to consider (default 1)
    :type Mw_max: float
    :param Mw_spacing: resolution of the moment magnitude search (default 0.1)
    :type Mw_spacing: float
    :param vp: P-wave velocity at the source location in m/s (default 5000.)
    :type vp: float
    :param vs: S-wave velocity at the source location in m/s (default 3500.)
    :type vs: float
    :param rho: Density at the source location in kg/m**3 (default 2400.)
    :type rho: float
    :param STALTA_threshold: STA/LTA threshold for a trigger to be declared
    (default 3)
    :type STALTA_threshold: float
    :param SSD: Static stress drop (default 0.01)
    :type SSD: float
    :param noise_level: Noise amplitude level at sensor in the unit of
    sensor response (e.g. velocity or acceleration) (default 0.02 (m/s**2))
    :type noise_level: either float or numpy array with the same dimensions
    as sensor_loc
    :param acceleration: If true noise_level is assumed to be acceleration
    and calculation are performed consequently
    :type acceleration: bool
    :rparam: return the minimum matnigude that sensitivity in Miminum
    magnitude sensitivity
    :rtype: float

    .. note:

    """
    nMw = (Mw_max - Mw_min) / Mw_spacing + 1
    Mw = np.linspace(Mw_min, Mw_max, nMw)

    att = geometrical_spreading(raypath) * radiation_pattern_attenuation()[0]

    signal_energy = []

    for Mw_ in zip(Mw):
        st = synthetic_seismogram(Mw_, vp=vp, vs=vs, rho=rho, SSD=SSD,
                                  pwave=True, )
        if acceleration:
            amp = att * np.std(st.differentiate().differentiate().data)
        else:
            amp = att * np.std(st.differentiate())

        amp = np.std(att * st.data)
        signal_energy.append(amp ** 2)

    signal_energy = np.array(signal_energy)

    if np.any(np.abs(signal_energy / (noise_level ** 2)) > STALTA_threshold):
        return \
            Mw[np.abs(signal_energy / (noise_level ** 2)) > STALTA_threshold][
                0]
    else:
        return None


def moment_magnitude(stream, cat, inventory, vp, vs, only_triaxial=True,
                     density=2700, min_dist=20, win_length=0.04,
                     len_spectrum=2 ** 12, clipped_fraction=0.1,
                     max_frequency=600, preferred_origin_only=True):
    """
    WARNING
    Calculate the moment magnitude for an event.
    :param stream: seismogram
    :type stream: uquake.Stream # does not exist yet
    :param cat: catalog object
    :type cat: uquake.core.event.Catalog
    :param inventory: network information (contains stations information)
    :type inventory: uquake.station.Site
    :param vp: P-wave velocity
    :type vp: float or uquake.core.data.Grid
    :param vs: S-wave velocity
    :type vs: float or uquake.core.data.Grid
    :param only_triaxial: whether only triaxial sensor are used in the
    magnitude calculation (optional) (not yet implemented)
    :type only_triaxial: bool
    :param density: density in kg / m**3 (assuming homogeneous for now)
    :type density: float
    :param win_length: length of the window in second in which magnitude is
    calculated
    :type win_length: float
    :param min_dist: minimum distance between sensor an event to allow
    magnitude calculation
    :param len_spectrum: length of the spectrum
    :param clipped_fraction: allowed clipped fraction (fraction of the
    signal equal to the min or the max.
    :param max_frequency: maximum frequency used in the calculation on
    magnitude. After a certain frequency, the noise starts to dominate the
    signal and the biases the calculation of the magnitude and corner
    frequency.
    :param preferred_origin_only: calculates the magnitude for the
    preferred_origin only
    :rtype: uquake.core.event.Catalog
    """

    # rigidity in Pa (shear-wave modulus)

    if only_triaxial:
        logger.info(
            'only triaxial sensor will be used in magnitude calculation')

    fcs = []

    quality = {'station_code': [], 'phase': [], 'origin_id': [], 'quality': []}

    if preferred_origin_only:
        origins = [cat[0].preferred_origin()]

    else:
        origins = cat[0].origins

    for origin in origins:
        ev_loc = np.array([origin.x, origin.y, origin.z])

        if not ((type(vp) == np.float) or (type(vp) == np.int)):
            vp_src = vp.interpolate(ev_loc, grid_space=False)
            vs_src = vs.interpolate(ev_loc, grid_space=False)
        else:
            vp_src = vp
            vs_src = vs

        moment_magnitudes = []
        corner_frequencies = []
        stations = []

        spectrum_norm_matrix = []
        frequencies = []
        indices = []
        for k, arr in enumerate(origin.arrivals):
            pick = arr.get_pick()
            network_code = pick.waveform_id.network_code
            station_code = pick.waveform_id.station_code
            location_code = pick.waveform_id.location_code
            travel_time = arr.get_pick().time - origin.time
            # ensuring backward compatibility
            if not pick:
                pick = cat[0].picks[k]
            at = pick.time
            phase = pick.phase_hint

            sensor_response = inventory.select(network=network_code,
                                               station=station_code,
                                               location=location_code)

            if sensor_response is None:
                logger.warning(f'no response was found in the inventory for '
                               f'sensor '
                               f'{network_code}.{station_code}.'
                               f'{location_code}')
                continue

            if sensor_response[0][0][0].response is None:
                logger.warning(f'no response was found in the inventory for '
                               f'sensor '
                               f'{network_code}.{station_code}.'
                               f'{location_code}')
                continue

            st_loc = sensor_response[0][0][0].loc
            if not sensor_response:
                logger.warning(f'sensor response not found for sensor '
                               f'{network_code}.{station_code}'
                               f'.{location_code}')
                continue

            poles = np.abs(sensor_response[0][0][0].response.get_paz().poles)
            st_trs = stream.select(network=network_code,
                                   station=station_code,
                                   location=location_code)

            if len(st_trs) == 0:
                continue

            if only_triaxial and (len(st_trs) < 3):
                continue

            # creating displacement pulse
            st_trs.detrend('demean').detrend('linear')
            st_trs.taper(max_percentage=0.05, type='cosine')

            data = st_trs.composite()[0].data

            len_max = len(data[data == np.max(data)]) + \
                      len(data[data == np.min(data)])

            if len_max / len(data) > clipped_fraction:
                logger.info('Clipped waveform detected: station %s '
                            'will not be used for magnitude calculation' %
                            sensor_response.code)
                continue

            pulse = st_trs.copy()
            pulse.attach_response(inventory)

            # filter the pulse using the corner frequency of the sensor

            sensor_min_freq = np.min(poles) / (2 * np.pi)
            window_min_freq = 1 / win_length

            low_bp_freq = np.max([sensor_min_freq, window_min_freq])
            high_bp_freq = np.max(poles) / (2 * np.pi)
            # if high_bp_freq > pulse[0].stats.sampling_rate / 2:
            #     high_bp_freq = pulse[0].stats.sampling_rate / 2.5

            high_bp_freq = max_frequency
            pulse = pulse.taper(max_percentage=0.05, type='cosine')
            # pulse.filter('bandpass', freqmin=low_bp_freq, freqmax=high_bp_freq)
            low_bp_freq1 = 10 ** (np.log10(low_bp_freq) - 0.2)
            low_bp_freq2 = low_bp_freq
            high_bp_freq1 = high_bp_freq
            high_bp_freq2 = 10 ** (np.log10(high_bp_freq) + 0.2)
            pre_filt = [low_bp_freq1, low_bp_freq2, high_bp_freq1,
                        high_bp_freq2]
            pulse.attach_response(inventory)
            # try:
            dp_trs = []
            for tr in pulse:
                try:
                    dp_trs = tr.remove_response(inventory=inventory,
                                                output='DISP',
                                                pre_filt=pre_filt)
                except Exception as e:
                    logger.error(e)

            dp = Stream(traces=dp_trs)

            # creating a signal containing only one for comparison
            tr_one = Trace(data=np.ones(len(pulse[0].data)))
            tr_one.stats = pulse[0].stats
            st_one = Stream(traces=[tr_one])

            dp = dp.trim(starttime=at - 0.01, endtime=at + 2 * win_length)
            dp = dp.taper(type='cosine', max_percentage=0.5, max_length=0.08,
                          side='left')

            # applying the same operation to the one signal
            # st_one_trimmed = st_one.trim(starttime=at - 0.01,
            #                              endtime=at + 2 * win_length)
            # st_one_taper = st_one_trimmed.taper(type='cosine',
            #                                     max_percentage=0.5,
            #                                     max_length=0.08,
            #                                     side='left')
            #
            dp_spectrum = np.zeros(len_spectrum)
            water_level = 1e-15
            for tr in dp:
                dp_spectrum += np.abs(np.fft.fft(tr.data, n=len_spectrum))
            # one_spectrum = np.fft.fft(st_one_taper[0].data, n=len_spectrum)

            # dp_spectrum_scaled = dp_spectrum # / (one_spectrum + water_level)

            if arr.distance is not None:
                hypo_dist = arr.distance
            else:
                hypo_dist = np.linalg.norm(st_loc - ev_loc)

            if hypo_dist < min_dist:
                continue

            radiation = radiation_pattern_attenuation()
            if phase.lower() == 'P':
                radiation = radiation[0]
                v_src = vp_src
            else:
                radiation = radiation[1]
                v_src = vs_src

            sr = dp[0].stats.sampling_rate

            f = np.fft.fftfreq(len_spectrum, 1 / sr)

            anelastic = np.exp(travel_time)

            spectrum_norm = dp_spectrum / radiation * hypo_dist * 4 * \
                            np.pi * density * v_src ** 3 / sr * anelastic

            fi = np.nonzero((f >= low_bp_freq) & (f <= high_bp_freq))[0]
            # fr = np.nonzero((f < low_bp_freq) | (f > high_bp_freq))[0]
            # spectrum_norm[fr] = np.nan
            spectrum_norm_matrix.append(np.abs(spectrum_norm[fi]))
            frequencies.append(f[fi])
            indices.append(np.ones(len(fi)) * k)

        if not spectrum_norm_matrix:
            continue

        st_count = len(spectrum_norm_matrix)
        spectrum_norm = np.nanmedian(spectrum_norm_matrix, axis=0)
        f = np.median(frequencies, axis=0)
        fi = np.nonzero((np.isnan(spectrum_norm) == False) & (f > 0))[0]

        p_opt, p_cov = curve_fit(spectral_function, f[fi],
                                 np.log10(spectrum_norm[fi]),
                                 (10, 100, 100), bounds=((1, 0, 10),
                                                         (100, 1000, 5000)))

        mw = 2 / 3.0 * p_opt[0] - 6.02
        mu = 29.5e9
        dnw1 = 2 / 3.0 * (p_opt[0] - p_cov[0, 0] / 2) - 6.02
        dnw2 = 2 / 3.0 * (p_opt[0] + p_cov[0, 0] / 2) - 6.02
        dmw = dnw2 - dnw1
        fc = p_opt[1]

        mag = event.Magnitude(mag=mw,
                              station_count=st_count, magnitude_type='Mw',
                              evaluation_mode=origin.evaluation_mode,
                              evaluation_status=origin.evaluation_status,
                              origin_id=origin.resource_id)

        mag.corner_frequency_hz = fc

        from obspy.core.event import QuantityError
        mag.mag_errors = QuantityError(uncertainty=dmw)
        # mag.fc_errors = QuantityError(uncertainty=dfc)
        cat[0].magnitudes.append(mag)
        if origin.resource_id == cat[0].preferred_origin().resource_id:
            cat[0].preferred_magnitude_id = mag.resource_id

    return cat




import matplotlib.pyplot as plt
import numpy as np
from obspy.core.event.base import Comment
from obspy.core.event.source import FocalMechanism, NodalPlane, NodalPlanes
from obspy.imaging.beachball import aux_plane

from hashwrap.hashwrapper import calc_focal_mechanisms
from .logging import logger


def calc(cat, settings):
    """
    Prepare input arrays needed to calculate focal mechanisms
    and pass these into hashwrap.hashwrapper

    Return list of obspy focalmechanisms & list of matplotlib figs

    :param cat: obspy.core.event.Catalog
    :type list: list of obspy.core.event.Events or microquake.core.event.Events
    :param settings:hash settings
    :type settings dictionary

    :returns: obsy_focal_mechanisms, matplotlib_figures
    :rtype: list, list
    """

    fname = 'calc_focal_mechanism'

    plot_focal_mechs = settings.plot_focal_mechs

    sname = []
    p_pol = []
    p_qual = []
    qdist = []
    qazi = []
    qthe = []
    sazi = []
    sthe = []

    events = []

    for event in cat:

        event_dict = {}

        origin = event.preferred_origin()

        event_dict['event_info'] = origin.time.datetime.strftime('%Y-%m-%d '
                                                                 '%H:%M:%S')
        event_dict['event'] = {}
        event_dict['event']['qdep'] = origin.loc[2]
        event_dict['event']['sez'] = 10.
        event_dict['event']['icusp'] = 1234567

        arrivals = [arr for arr in event.preferred_origin().arrivals if
                    arr.phase == 'P']

        for arr in arrivals:

            if not arr.get_pick():
                logger.warning(
                    f"Missing pick for arrival {arr.resource_id} on"
                    f" event {event.resource_id}")
                continue

            if arr.get_pick().snr is None:
                logger.warning("%s P arr pulse_snr == NONE !!!" %
                               arr.pick_id.get_referred_object(
                               ).waveform_id.station_code)
                continue

            if arr.polarity is None:
                continue

            sname.append(arr.pick_id.get_referred_object(
            ).waveform_id.station_code)
            p_pol.append(arr.polarity)
            qdist.append(arr.distance)
            qazi.append(arr.azimuth)
    # MTH: both HASH and test_stereo expect takeoff theta measured wrt
            # vertical Up!
            qthe.append(180. - arr.takeoff_angle)
            sazi.append(2.)
            sthe.append(10.)

            if arr.get_pick().snr <= 6:
                qual = 0
            else:
                qual = 1
            p_qual.append(qual)

        event_dict['sname'] = sname
        event_dict['p_pol'] = p_pol
        event_dict['p_qual'] = p_qual
        event_dict['qdist'] = qdist
        event_dict['qazi'] = qazi
        event_dict['qthe'] = qthe
        event_dict['sazi'] = sazi
        event_dict['sthe'] = sthe

    events.append(event_dict)

    outputs = calc_focal_mechanisms(events, settings,
                                    phase_format='FPFIT')

    focal_mechanisms = []

    plot_figures = []

    for i, out in enumerate(outputs):
        logger.info("%s.%s: Process Focal Mech i=%d" % (__name__, fname, i))
        p1 = NodalPlane(strike=out['strike'], dip=out['dip'], rake=out['rake'])
        s, d, r = aux_plane(out['strike'], out['dip'], out['rake'])
        p2 = NodalPlane(strike=s, dip=d, rake=r)

        fc = FocalMechanism(nodal_planes=NodalPlanes(nodal_plane_1=p1,
                                                     nodal_plane_2=p2),
                            azimuthal_gap=out['azim_gap'],
                            station_polarity_count=out[
                                'station_polarity_count'],
                            station_distribution_ratio=out['stdr'],
                            misfit=out['misfit'],
                            evaluation_mode='automatic',
                            evaluation_status='preliminary',
                            comments=[Comment(text="HASH v1.2 Quality=[%s]"
                                                   % out['quality'])]
                            )

        focal_mechanisms.append(fc)

        event = events[i]

        title = "%s (s,d,r)_1=(%.1f,%.1f,%.1f) _2=(%.1f,%.1f,%.1f)" % \
                (event['event_info'], p1.strike, p1.dip, p1.rake, p2.strike,
                 p2.dip, p2.rake)

        if plot_focal_mechs:
            gcf = test_stereo(np.array(event['qazi']),
                              np.array(event['qthe']),
                              np.array(event['p_pol']),
                              sdr=[p1.strike, p1.dip, p1.rake],
                              event_info=event['event_info'])
            # sdr=[p1.strike,p1.dip,p1.rake], title=title)
            plot_figures.append(gcf)

    return focal_mechanisms, plot_figures


def test_stereo(azimuths, takeoffs, polarities, sdr=[], event_info=None):
    '''
        Plots points with given azimuths, takeoff angles, and
        polarities on a stereonet. Will also plot both planes
        of a double-couple given a strike/dip/rake
    '''

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='stereonet')
    up = polarities > 0
    dn = polarities < 0

    plot_upper_hemisphere = True

# This assumes takeoffs are measured wrt vertical up
#   so that i=0 (vertical) up has plunge=90:
#          ax.line(plunge, trend) - where plunge=90 plots at center and
    #          plunge=0 at edge
    h_rk = ax.line(90.-takeoffs[up], azimuths[up], 'bo')  # compressional
    # first arrivals
    h_rk = ax.line(90.-takeoffs[dn], azimuths[dn], 'go', fillstyle='none')

    if sdr:
        s1, d1, r1 = sdr[0], sdr[1], sdr[2]
        s2, d2, r2 = aux_plane(*sdr)

        if plot_upper_hemisphere:
            s1 += 180.
            s2 += 180.
        h_rk = ax.plane(s1, d1, 'g')
        ax.pole(s1, d1, 'gs', markersize=7)
        h_rk = ax.plane(s2, d2, 'b')
        ax.pole(s2, d2, 'bs', markersize=7)

    ax.grid(True)

    plt.title("upper hemisphere")

    title = event_info + " (s,d,r)=(%.1f, %.1f, %.1f)" % (s1, d1, r1)

    if title:
        plt.suptitle(title)

    # plt.show()

    return plt.gcf()
